import copy
import numpy as np
import pathlib
import pickle as pkl

from loguru import logger
from matplotlib.axes import Axes
from PyQt5 import Qt
from typing import List, Optional, Union, Type

from .window_base import MainWindow, SceneDictType
from .utils.types import contains_scene


__all__ = ['display_dataset', 'display_dataset_custom', 'MainWindowDataset']


class MainWindowDataset(MainWindow):

    def __init__(
        self,
        scenes: List[SceneDictType],
        fps: float,
        video_directory: Optional[Union[pathlib.Path, str]] = None,
        horizontal: bool = True,
        loop: bool = True,
        show_labels: bool = False,
        show_frame_index: bool = False,
        use_scene_cam: bool = False,
        store_directory: Optional[Union[pathlib.Path, str]] = None,
        debug: bool = False,
        parent: Qt.QWidget = None
    ):
        self.scenes = scenes
        self.fps = fps
        if len(scenes) == 0:
            return
        self.__frame_idx = 0
        self.__show_frame_index = show_frame_index
        scene_dict = self.scenes[0]

        image_keys = [key for key, value in scene_dict.items() if isinstance(value, np.ndarray)]
        scene_keys = [key for key, value in scene_dict.items() if contains_scene(value)]
        figure_keys = [key for key, value in scene_dict.items() if isinstance(value, Axes)]
        logger.debug(f"Image keys: {image_keys}")
        logger.debug(f"Scene keys: {scene_keys}")
        logger.debug(f"Figure keys: {figure_keys}")

        super(MainWindowDataset, self).__init__(
            image_keys=image_keys,
            scene_keys=scene_keys,
            figure_keys=figure_keys,
            video_directory=video_directory,
            horizontal=horizontal,
            show_labels=show_labels,
            use_scene_cam=use_scene_cam,
            debug=debug,
            parent=parent
        )
        title = self.__make_title()
        self.render_(scene_dict, resetcam=True, title=title)

        # If store_directory is given, save the scene list to the directory.
        if store_directory is not None:
            logger.info(f"Storing scene in directory {store_directory}")
            store_directory = pathlib.Path(store_directory)
            if store_directory.exists():
                logger.info(f"Directory {store_directory} already exists, overwriting...")
            store_directory.mkdir(parents=True, exist_ok=True)
            for k, scene in enumerate(scenes):
                pkl_file = store_directory / f"{k:05d}.pkl"
                with open(pkl_file, 'wb') as f:
                    pkl.dump(scene, f, protocol=pkl.HIGHEST_PROTOCOL)
            logger.info("Storing scenes done")

        # Setup looping timer and connect it to render a new scene at every timer callback.
        # If loop is true, start looping right away.
        self.timer = Qt.QTimer(self)
        self.timer.timeout.connect(self.render_loop)  # noqa
        if loop:
            self.toggle_looping()

    def render_loop(self):
        self.render_current_index()
        self.__frame_idx += 1  # TODO: run forward or backward

    def render_current_index(self, resetcam: bool = False) -> None:
        if 0 <= self.__frame_idx < len(self.scenes):
            scene_dict = self.scenes[self.__frame_idx]

            # The same matplotlib.Axes cannot be used multiple times, as it is assigned to a figure during
            # plotting. Therefore, we need to create a copy of the scene dictionary to be able to reuse it
            # when plotting the scene again later.
            if any(isinstance(value, Axes) for value in scene_dict.values()):
                scene_dict_safe = dict()
                for key, value in scene_dict.items():
                    if key in self.figure_key_dict.keys():
                        value = copy.deepcopy(value)
                    scene_dict_safe[key] = value
            else:
                scene_dict_safe = scene_dict

            title = self.__make_title()
            self.render_(scene_dict_safe, resetcam=resetcam, title=title)
        else:
            # Reset frame index to a valid index.
            self.__frame_idx = max(min(self.__frame_idx, len(self.scenes) - 1), 0)
            # Stop the timer, if it is running.
            if self.timer.isActive():
                self.timer.stop()

    def _on_key(self, event_dict) -> None:
        super(MainWindowDataset, self)._on_key(event_dict)
        key_pressed = event_dict["keypress"]
        if key_pressed == "s":
            self.toggle_looping()
        elif key_pressed == "n" and not self.timer.isActive():
            self.__frame_idx += 1
            self.render_current_index()
        elif key_pressed == "b" and not self.timer.isActive():
            self.__frame_idx -= 1
            self.render_current_index()

    def __make_title(self) -> Optional[str]:
        if not self.__show_frame_index:
            return None
        return f"Frame {self.__frame_idx + 1}/{len(self.scenes)}"

    @staticmethod
    def print_usage():
        MainWindow.print_usage()
        print("\ts: play / pause",
              "\n\tn: next",
              "\n\tb: previous")

    def toggle_looping(self):
        if self.timer.isActive():
            self.timer.stop()
        else:
            self.timer.start(int(1 / self.fps * 1000))  # in milliseconds


def display_dataset_custom(
    main_window_class: Type[MainWindowDataset],
    scenes: List[SceneDictType],
    horizontal: bool = True,
    loop: bool = False,
    fps: float = 30.0,
    video_directory: Optional[Union[pathlib.Path, str]] = None,
    show_labels: bool = False,
    show_frame_index: bool = False,
    use_scene_cam: bool = False,
    store_directory: Optional[Union[pathlib.Path, str]] = None,
    debug: bool = False
):
    """Display scenes stored in scene iterator as PyQt app with a custom main window dataset class.

    The scene iterator yields a dictionary that describes the elements of the scene, one dictionary per frame.
    See SceneDictType for currently supported object types.

    Args:
        main_window_class
        scenes: sorted list of scene dictionaries.
        horizontal: window orientation, horizontal or vertical stacking.
        loop: start looping from beginning.
        fps: frames (i.e. scenes) per second for looping.
        video_directory: directory for storing screenshots.
        show_labels: display the scene dict keys.
        show_frame_index: display the current frame index.
        use_scene_cam: use camera transform from trimesh scenes.
        store_directory: path to directory storing scene pickle files.
        debug: printing debug information.
    """
    if len(scenes) == 0:
        logger.warning("No scenes to display.")
        return

    app = Qt.QApplication([])
    window = main_window_class(
        scenes=scenes,
        fps=fps,
        video_directory=video_directory,
        horizontal=horizontal,
        loop=loop,
        show_labels=show_labels,
        show_frame_index=show_frame_index,
        debug=debug,
        use_scene_cam=use_scene_cam,
        store_directory=store_directory
    )
    app.aboutToQuit.connect(window.on_close)
    app.exec_()


def display_dataset(
    scenes: List[SceneDictType],
    horizontal: bool = True,
    loop: bool = False,
    fps: float = 30.0,
    video_directory: Optional[Union[pathlib.Path, str]] = None,
    show_labels: bool = False,
    show_frame_index: bool = False,
    use_scene_cam: bool = False,
    store_directory: Optional[Union[pathlib.Path, str]] = None,
    debug: bool = False
):
    """Display scenes stored in scene iterator as PyQt app.

    The scene iterator yields a dictionary that describes the elements of the scene, one dictionary per frame.
    See SceneDictType for currently supported object types.

    Args:
        scenes: sorted list of scene dictionaries.
        horizontal: window orientation, horizontal or vertical stacking.
        loop: start looping from beginning.
        fps: frames (i.e. scenes) per second for looping.
        video_directory: directory for storing screenshots.
        show_labels: display the scene dict keys.
        show_frame_index: display the current frame index.
        use_scene_cam: use camera transform from trimesh scenes.
        store_directory: path to directory storing scene pickle files.
        debug: printing debug information.
    """
    return display_dataset_custom(
        main_window_class=MainWindowDataset,
        scenes=scenes,
        horizontal=horizontal,
        loop=loop,
        fps=fps,
        video_directory=video_directory,
        show_labels=show_labels,
        show_frame_index=show_frame_index,
        use_scene_cam=use_scene_cam,
        store_directory=store_directory,
        debug=debug
    )



import balanna.camera_trajectories as camera_trajectories
import balanna.components as components

from balanna.window_dataset import display_dataset, display_dataset_custom, MainWindowDataset
from balanna.window_generator import display_generated, display_scenes, display_generated_custom, MainWindowGenerator
from balanna.window_base import SceneDictType


__all__ = [
  "camera_trajectories", 
  "components",
  "display_dataset",
  "display_dataset_custom",
  "display_scenes",
  "display_generated",
  "display_generated_custom",
  "MainWindowDataset",
  "MainWindowGenerator",
  "SceneDictType",
]



import os
import numpy as np
import trimesh
import trimesh.creation

from loguru import logger
from scipy.spatial.transform import Rotation
from typing import List, Optional, Tuple, Union


__all__ = [
    "show_box_w_transform",
    "show_box_w_aabb",
    "show_ellipsoid",
    "show_plane",
    "show_mesh",
    "show_trimesh",
    "show_grid",
    "show_point_cloud",
    "show_trajectory",
    "show_camera",
    "show_axis",
    "show_axes",
    "show_sphere",
    "show_capsule",
    "show_cylinder",
]


RGBType = Tuple[int, int, int]
RGBAType = Tuple[int, int, int, int]
RGBorRGBAType = Union[RGBAType, RGBType]


def show_box_w_transform(
    extents: Tuple[float, float, float],
    transform: np.ndarray,
    color: RGBorRGBAType = (120, 120, 120),
    name: Optional[str] = None,
    scene: Optional[trimesh.Scene] = None,
):
    """Create a 3D box with the given transformation matrix to the plane coordinate frame (normal = z axis).

    Args:
        extents: box extents (x, y, z).
        normal: plane normal vector (3).
        point: point on the plane (3).
        color: box color as RGB / RGBA tuple in 0...255 (3/4).
        name: scene node name of mesh.
        scene: scene to add mesh to, if None a new scene will be created.
    """
    if len(extents) != 3:
        raise ValueError(f"Invalid extents, expected (3,), got {len(extents)}")
    if transform.shape != (4, 4):
        raise ValueError(f"Expected transformation matrix of shape (4, 4), got {transform.shape}")

    if scene is None:
        scene = trimesh.Scene()
    color_uint8 = np.array(color, dtype=np.uint8)

    box = trimesh.creation.box(extents=extents, transform=transform, name=name)
    box.visual.face_colors = np.repeat(color_uint8[None, :], len(box.faces), axis=0)

    scene.add_geometry(box)
    return scene


def show_plane(
    normal: np.ndarray,
    point: np.ndarray,
    extent_xy: float = 5.0,
    extent_z: float = 0.01,
    color: RGBorRGBAType = (120, 120, 120),
    name: Optional[str] = None,
    scene: Optional[trimesh.Scene] = None,
):
    """Create a 3D plane as a flat 3D box with the normal vector and a point on the plane.

    Args:
        normal: plane normal vector (3).
        point: point on the plane (3).
        extent_xy: plane extent in x and y direction.
        extent_z: plane extent in z direction (thickness of the plane box).
        name: scene node name of mesh.
        scene: scene to add mesh to, if None a new scene will be created.
    """
    if normal.shape != (3,):
        raise ValueError(f"Invalid normal shape, expected (3,), got {normal.shape}")
    if point.shape != (3,):
        raise ValueError(f"Invalid point shape, expected (3,), got {point.shape}")
    unit_axis = np.array([0, 0, 1])  # normal of the unit plane in trimesh

    # Construct transform from normal vector and point on plane. The plane normal is aligned with the z-axis.
    # Therefore, the rotation matrix R aligns the unit z-axis with the plane normal.
    R, _ = Rotation.align_vectors(normal[None], unit_axis[None])
    T = np.eye(4)
    T[:3, :3] = R.as_matrix()
    T[:3, 3] = point

    extents = (extent_xy, extent_xy, extent_z)
    return show_box_w_transform(extents, transform=T, name=name, color=color, scene=scene)


def show_box_w_aabb(
    aabb: np.ndarray, name: Optional[str] = None, scene: Optional[trimesh.Scene] = None
):
    """Create a 3D box with the min and max corner.

    Args:
        aabb: [(x_min, y_min, z_min), (x_max, y_max, z_max)] (2, 3).
        name: scene node name of mesh.
        scene: scene to add mesh to, if None a new scene will be created.
    """
    if aabb.shape != (2, 3):
        raise ValueError(f"Invalid aabb shape, expected (2, 3), got {aabb.shape}")

    if scene is None:
        scene = trimesh.Scene()

    box = trimesh.creation.box(bounds=aabb, name=name)
    scene.add_geometry(box)
    return scene


def show_mesh(
    vertices: np.ndarray,
    faces: np.ndarray,
    transform: np.ndarray = np.eye(4),
    name: Optional[str] = None,
    scene: Optional[trimesh.Scene] = None,
    face_color: Optional[np.ndarray] = None,
    vertex_color: Optional[np.ndarray] = None,
) -> trimesh.Scene:
    """Create a mesh from vertices and faces and add it to the scene.

    Args:
        vertices: mesh vertices (N, 3).
        faces: mesh faces.
        transform: transform mesh before adding it to scene (4, 4).
        name: scene node name of mesh.
        scene: scene to add mesh to, if None a new scene will be created.
        face_color: face colors (F, 3).
        vertex_color: vertex colors (N, 3/4).
    """
    if vertex_color is not None and vertex_color.shape[0] != vertices.shape[0]:
        raise ValueError(
            f"Invalid vertex colors, must be {vertices.shape}, got {vertex_color.shape}"
        )
    if vertex_color is not None and face_color is not None:
        raise ValueError("Cannot set both vertex and face color")

    mesh = trimesh.Trimesh(vertices, faces)
    if face_color is not None:
        mesh.visual.face_colors[:, :3] = face_color
    elif vertex_color is not None:
        mesh.visual.vertex_colors = vertex_color
    else:
        mesh.visual.face_colors[:, :3] = [224, 120, 120]

    return show_trimesh(mesh, name, transform, scene, copy_mesh=False)


def show_trimesh(
    mesh: trimesh.Trimesh,
    name: Optional[str] = None,
    transform: np.ndarray = np.eye(4),
    scene: Optional[trimesh.Scene] = None,
    copy_mesh: bool = True,
):
    """Add a trimesh mesh to the scene.

    Args:
        mesh: trimesh mesh to add.
        name: scene node name of mesh.
        transform: transform mesh before adding it to scene (4, 4).
        scene: scene to add mesh to, if None a new scene will be created.
        copy_mesh: copy the mesh before adding it to the scene and applying the transform.
    """
    if scene is None:
        scene = trimesh.Scene()
    mesh_scene = mesh.copy() if copy_mesh else mesh
    mesh_scene = mesh_scene.apply_transform(transform)
    scene.add_geometry(mesh_scene, node_name=name)
    return scene


def show_grid(
    xy_min: float = -10.0,
    xy_max: float = 10.0,
    z: float = 0.0,
    resolution: float = 1.0,
    alpha: float = 255,
    dark_color: RGBAType = (120, 120, 120),
    light_color: RGBAType = (255, 255, 255),
    transform: np.ndarray = np.eye(4),
    scene: trimesh.Scene = None,
) -> trimesh.Scene:
    """Create a 2D squared grid in the xy-plane, color it in a chessboard pattern and transform it accordingly.

    Args:
        xy_min: minimal coordinate in xy direction, i.e. (xy_min, xy_min).
        xy_max: maximal coordinate in xy direction, i.e. (xy_max, xy_max).
        z: plane height at creation.
        resolution: cell size.
        alpha: color alpha value for all cells.
        dark_color: chessboard dark color.
        light_color: chessboard light color.
        transform: mesh transform after creation (in relation to center point (0, 0, z)).
        scene: scene to add mesh to, if None a new scene will be created.
    """
    if scene is None:
        scene = trimesh.Scene()

    num_points_per_row = int((xy_max - xy_min) // resolution) + 1
    w = num_points_per_row - 1
    num_faces = w**2

    # Compute vertices as 2D mesh-grid in the xy-plane. Set all z components to the given z value.
    vertices_x, vertices_y = np.meshgrid(
        np.linspace(xy_min, xy_max, num_points_per_row),
        np.linspace(xy_min, xy_max, num_points_per_row),
    )
    vertices_x = vertices_x.flatten()
    vertices_y = vertices_y.flatten()
    vertices = np.stack([vertices_x, vertices_y, np.ones_like(vertices_x) * z], axis=-1)

    # Compute grid faces and face colors to create chessboard pattern.
    faces = []
    face_colors = np.zeros((num_faces, 4), dtype=np.uint8)
    face_colors[:, -1] = alpha  # alpha value
    sym = [light_color, dark_color]
    face_idx = 0
    for i in range(w):
        for j in range(w):
            # add face in counter-clockwise vertex order.
            face_ij = (
                i * num_points_per_row + j,
                i * num_points_per_row + j + 1,
                (i + 1) * num_points_per_row + j + 1,
                (i + 1) * num_points_per_row + j,
            )
            faces.append(face_ij)

            # if even number of elements in row: at beginning of row, swap colors.
            # if odd: order is automatically swapped after finishing one row.
            if face_idx % w == 0 and w % 2 == 0:
                sym.reverse()
            idx = 0 if face_idx % 2 == 0 else 1  # alternate index for chessboard pattern
            face_colors[face_idx, :3] = sym[idx]

            # Increment face index.
            face_idx += 1
    face_colors = np.concatenate([face_colors, face_colors])

    grid = trimesh.Trimesh(vertices, faces, face_colors=face_colors)
    scene.add_geometry(grid, transform=transform)
    return scene


def show_point_cloud(
    vertices: np.ndarray,
    transform: np.ndarray = np.eye(4),
    colors: Optional[np.ndarray] = None,
    name: Optional[str] = None,
    point_size: float = 4,
    scene: Optional[trimesh.Scene] = None,
) -> trimesh.Scene:
    """Add vertices as trimesh point cloud to the scene.

    Trimesh expects a flat point cloud of shape (N, 3). Therefore, if the point cloud is not flat, it will be
    flattened before passing it to trimesh.

    Args:
        vertices: point cloud to visualize (..., 3).
        transform: transform point cloud before adding it to scene (4, 4).
        colors: RGB color for each point (..., 3) as uint8 color 0...255..
        name: scene node name of point cloud mesh.
        point_size: point size.
        scene: scene to add point cloud mesh to, if None a new scene will be created.
    """
    if scene is None:
        scene = trimesh.Scene()
    if transform.shape != (4, 4):
        raise ValueError(f"Expected single transformation matrix, got {transform.shape}")

    vertices_flat = vertices.reshape(-1, 3)
    if colors is None:
        colors = np.zeros((len(vertices_flat), 3))
        colors[:, 0] = 255  # default = red
    pc = trimesh.PointCloud(vertices_flat, colors=colors, metadata={"point_size": point_size})
    scene.add_geometry(pc, transform=transform, node_name=name)
    return scene


def show_trajectory(
    trajectory: np.ndarray,
    colors: Union[RGBorRGBAType, np.ndarray] = (255, 0, 0),
    fade_out: bool = False,
    scene: Optional[trimesh.Scene] = None,
) -> trimesh.Scene:
    """Add a colored trajectory to the scene

    Args:
        trajectory: path to be drawn (N, 3).
        colors: path color (tuple for uniformly colored line as RGB(-A) tuple & array for segment-wise color).
        alpha: path alpha value. If fade_out, max alpha value.
        fade_out: fade out line by linearly decreasing the alpha value along the path length.
        scene: scene to add path to, if None a new scene will be created.
    """
    if len(trajectory.shape) != 2 or trajectory.shape[-1] != 3:
        raise ValueError(f"Invalid size of trajectory, expected (N, 3), got {trajectory.shape}")
    if scene is None:
        scene = trimesh.Scene()

    # If the trajectory just contains a single point, then it cannot be drawn. Just return the scene.
    if trajectory.shape[0] < 2:
        logger.warning("Trajectory must contain at least two points, skipping ...")
        return scene

    # In trimesh a path consists of line segments, each connecting one vertex with the next one. The
    # way of connection is determined from the entities (similar to faces in 3D meshes).
    # Each line segment could have a different color, but vedo just supports uniformly colored lines,
    # which would be supported, but over-complicate things here.
    # process = False, otherwise vertices are weirdly reconnected
    num_points = trajectory.shape[0]
    entities = [trimesh.path.entities.Line([j, j + 1]) for j in range(num_points - 1)]

    if isinstance(colors, tuple):
        alpha = colors[-1] if len(colors) == 4 else 255
        colors_rgb = colors[:3]

        alphas = [
            alpha * ((k + 2) / num_points) if fade_out else alpha for k in range(num_points - 1)
        ]
        entity_colors = [(*colors_rgb, alphas[k]) for k in range(num_points - 1)]

    elif isinstance(colors, np.ndarray):
        if colors.dtype != np.uint8:
            raise ValueError(f"Invalid colors dtype, expected uint8, got {colors.dtype}")
        if len(colors) != num_points - 1:
            raise ValueError(f"Invalid colors length, expected {num_points - 1}, got {len(colors)}")
        if len(colors.shape) != 2:
            raise ValueError(f"Invalid colors shape, expected (N-1, 3/4), got {colors.shape}")

        if colors.shape[1] == 3:
            colors_rgb = colors
            alphas = [
                255 * (k + 2) / num_points if fade_out else 255 for k in range(num_points - 1)
            ]
        elif colors.shape[1] == 4:
            if fade_out:
                raise ValueError("Fade out not supported for RGBA colors")
            colors_rgb = colors[:, :3]
            alphas = colors[:, 3]
        else:
            raise ValueError(f"Invalid colors shape, expected (N-1, 3/4), got {colors.shape}")

        entity_colors = [(*colors_rgb[k], alphas[k]) for k in range(num_points - 1)]

    # The vertices must be copied, otherwise the original trajectory might be overwritten afterwards.
    path = trimesh.path.Path3D(
        entities=entities,
        vertices=trajectory.copy(),
        colors=entity_colors,
        process=False,
    )
    scene.add_geometry(path)
    return scene


def show_camera(
    transform: np.ndarray,
    name: Optional[str] = None,
    scene: Optional[trimesh.Scene] = None,
) -> trimesh.Scene:
    """Load camera mesh and transform it to given pose.

    Args:
        transform: camera pose as transformation matrix (4, 4).
        name: scene node name of the camera mesh.
        scene: scene to add camera to, if None a new scene will be created.
    """
    if scene is None:
        scene = trimesh.Scene()
    if transform.shape != (4, 4):
        raise ValueError(f"Expected single transformation matrix, got {transform.shape}")
    script_path = os.path.basename(os.path.dirname(__file__))
    cam_mesh_path = os.path.join(script_path, "..", "meshes", "cam_z.obj")
    cam_mesh = trimesh.load_mesh(
        cam_mesh_path
    )  # already in camera orientation (z forward, x right, y down)
    cam_mesh = cam_mesh.apply_transform(transform)
    scene.add_geometry(cam_mesh, node_name=name)
    return scene


def show_axis(
    transform: np.ndarray,
    name: Optional[str] = None,
    size: float = 0.06,
    scene: Optional[trimesh.Scene] = None,
) -> trimesh.Scene:
    """Add coordinate axis as trimesh mesh to the scene.

    Args:
        transform: axis pose as transformation matrix (4, 4).
        name: scene node name of axis mesh.
        size: axis origin size and radius.
        scene: scene to add axis to, if None a new scene will be created.
    """
    if scene is None:
        scene = trimesh.Scene()
    if transform.shape != (4, 4):
        raise ValueError(f"Expected single transformation matrix, got {transform.shape}")
    axis_mesh = trimesh.creation.axis(size, transform=transform, axis_radius=size / 3)
    scene.add_geometry(axis_mesh, node_name=name)
    return scene


def show_axes(
    transforms: Union[List[np.ndarray], np.ndarray],
    sizes: Optional[List[float]] = None,
    scene: Optional[trimesh.Scene] = None,
):
    """Add coordinate axes for multiple frames to the scene.

    Args:
        transforms: axes poses as transformation matrices (N, 4, 4 or list of (4, 4)).
        sizes: axes sizes.
        scene: scene to add axis to, if None a new scene will be created.
    """
    if sizes is None:
        sizes = [0.04] * len(transforms)
    for tf, size in zip(transforms, sizes):
        scene = show_axis(tf, size=size, scene=scene)
    return scene


def show_sphere(
    center: np.ndarray,
    radius: float,
    color: RGBorRGBAType,
    scene: Optional[trimesh.Scene] = None,
    count: Optional[Tuple[int, int]] = None,
) -> trimesh.Scene:
    """Add a sphere to the scene.

    Args:
        center: sphere center coordinates (3).
        radius: sphere radius.
        color: RGB / RGBA sphere color in 0...255 (3/4).
        scene: scene to add sphere to.
        count: sphere resolution as number of latitude and longitude lines (2).
    """
    if scene is None:
        scene = trimesh.Scene()
    if radius < 0.01:
        raise ValueError(f"Invalid sphere radius {radius}, must be > 0.01")
    if count is not None and (count[0] < 3 or count[1] < 3):
        raise ValueError(f"Invalid sphere count {count}, must be >= 3")

    sphere = trimesh.creation.uv_sphere(radius=radius, count=count)
    color_uint8 = np.array(color, dtype=np.uint8)

    sphere.visual.face_colors = np.repeat(color_uint8[None, :], len(sphere.faces), axis=0)
    sphere.apply_translation(center)

    scene.add_geometry(sphere)
    return scene


def show_ellipsoid(
    center: np.ndarray,
    radii: np.ndarray,
    color: RGBorRGBAType,
    scene: Optional[trimesh.Scene] = None,
    count: Optional[Tuple[int, int]] = None,
) -> trimesh.Scene:
    """Add an ellipsoid to the scene.

    Args:
        center: ellipsoid center coordinates (3).
        radii: ellipsoid radii (3).
        color: RGB ellipsoid color in 0...255 (3).
        scene: scene to add ellipsoid to.
        count: ellipsoid resolution as number of latitude and longitude lines (2).
    """
    if scene is None:
        scene = trimesh.Scene()
    if np.any(radii < 0.01):
        raise ValueError(f"Invalid ellipsoid radii {radii}, must be > 0.01")
    if radii.shape != (3,):
        raise ValueError(f"Invalid ellipsoid radii shape, expected (3,), got {radii.shape}")
    if count is not None and (count[0] < 3 or count[1] < 3):
        raise ValueError(f"Invalid ellipsoid count {count}, must be >= 3")

    if count is None:
        count = (20, 20)
    nu, nv = count

    # Compute vertices for the ellipsoid using the ellipsoid parametric equation.
    vertices = np.empty(((nu + 1) * (nv + 1), 3))
    for i in range(nu + 1):
        theta = i * np.pi / nv  # angle for the latitude (0 to pi)
        for j in range(nv + 1):
            phi = j * 2 * np.pi / nv  # angle for the longitude (0 to 2*pi)
            x = radii[0] * np.sin(theta) * np.cos(phi)
            y = radii[1] * np.sin(theta) * np.sin(phi)
            z = radii[2] * np.cos(theta)
            vertices[i * (nv + 1) + j] = [x, y, z] + center

    # Compute faces for the ellipsoid.
    faces = np.empty((2 * nu * nv, 3), dtype=int)
    for i in range(nu):
        for j in range(nv):
            v1 = i * (nv + 1) + j
            v2 = (i + 1) * (nv + 1) + j
            v3 = (i + 1) * (nv + 1) + (j + 1)
            v4 = i * (nv + 1) + (j + 1)
            faces[2 * (i * nv + j)] = [v1, v2, v3]
            faces[2 * (i * nv + j) + 1] = [v1, v3, v4]

    mesh = trimesh.Trimesh(vertices=vertices, faces=faces)
    scene.add_geometry(mesh)
    return scene


def show_capsule(
    p1: np.ndarray,
    p2: np.ndarray,
    radius: float,
    color: RGBorRGBAType,
    scene: Optional[trimesh.Scene] = None,
    count: Optional[Tuple[int, int]] = None,
) -> trimesh.Scene:
    """Add a capsule to the scene.

    Args:
        p1: capsule start point (3).
        p2: capsule end point (3).
        radius: capsule radius.
        color: RGB capsule color in 0...255 (3).
        scene: scene to add capsule to.
        count: sphere resolution as number of latitude and longitude lines (2).
    """
    if scene is None:
        scene = trimesh.Scene()
    if radius < 0.01:
        raise ValueError(f"Invalid capsule radius {radius}, must be > 0.01")
    if np.allclose(p1, p2):
        raise ValueError(f"Invalid capsule points {p1}, {p2}, must be different")
    if count is not None and (count[0] < 3 or count[1] < 3):
        raise ValueError(f"Invalid sphere count {count}, must be >= 3")

    # The canonical capsule primitive is defined as a cylinder with its z-axis aligned with the world z-axis.
    # It is determined by a single point P in the center of the cylinder and a height h.
    # This function computes the transformation from this canonical capsule to the capsule defined here from the
    # points p1 and p2. The canonical representation is given by:
    # - The origin of the capsule (P = center).
    # - The height of the capsule (h).
    # - The radius of the capsule (r).
    # - The orientation from the canonical representation (Rc).
    dp = p2 - p1
    height = np.linalg.norm(dp)
    z_axis = np.array([0, 0, 1])
    dp_normed = dp / height

    # To compute the rotation matrix that rotates the z-axis to dp, we use the Rodrigues formula.
    # See https://math.stackexchange.com/questions/180418/calculate-rotation-matrix-to-align-vector-a-to-vector-b-in-3d
    v = np.cross(z_axis, dp_normed)
    s = np.linalg.norm(v)
    c = np.dot(z_axis, dp_normed)
    vx = np.array(
        [[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]]
    )  # skew-symmetric matrix of v

    Rc = np.eye(3) + vx + np.matmul(vx, vx) * (1 - c) / (s**2 + 1e-6)  # Rodrigues formula
    center = p1 + dp / 2  # equals P

    Tz = np.eye(4)
    Tz[:3, 3] = center
    Tz[:3, :3] = Rc

    capsule_mesh = trimesh.creation.capsule(radius=radius, height=height, transform=Tz, count=count)

    # Add color to capsule.
    color_uint8 = np.array(color, dtype=np.uint8)
    capsule_mesh.visual.face_colors = np.repeat(
        color_uint8[None, :], len(capsule_mesh.faces), axis=0
    )

    scene.add_geometry(capsule_mesh)
    return scene


def show_cylinder(
    T: np.ndarray,
    radius: float,
    height: float,
    color: RGBorRGBAType,
    scene: Optional[trimesh.Scene] = None,
    count: Optional[int] = None,
) -> trimesh.Scene:
    """Add a cylinder to the scene.

    Args:
        T: center pose as transformation matrix (4, 4).
        radius: cylinder radius.
        height: cylinder height (along body z-axis, height / 2 above and below origin).
        color: RGB capsule color in 0...255 (3).
        scene: scene to add capsule to.
        count: How many pie wedges should the cylinder have
    """
    if scene is None:
        scene = trimesh.Scene()
    if radius < 0.01:
        raise ValueError(f"Invalid cylinder radius {radius}, must be > 0.01")
    if height < 0.01:
        raise ValueError(f"Invalid cylinder height {height}, must be > 0.01")
    if count is not None and count < 3:
        raise ValueError(f"Invalid cylinder count {count}, must be >= 3")

    cylinder_mesh = trimesh.creation.cylinder(
        radius=radius, height=height, transform=T, count=count
    )

    # Add color to cylinder.
    color_uint8 = np.array(color, dtype=np.uint8)
    cylinder_mesh.visual.face_colors = np.repeat(
        color_uint8[None, :], len(cylinder_mesh.faces), axis=0
    )

    scene.add_geometry(cylinder_mesh)
    return scene



import argparse
import numpy as np
import pathlib
import pickle as pkl
import sys

from loguru import logger

from balanna.json_parser import load_scene_from_json
from balanna.components import show_point_cloud, show_grid
from balanna.window_dataset import display_dataset


def parse_arg_pairs(arg_pairs):
    splits = arg_pairs.split("=")
    if len(splits) != 2:
        raise argparse.ArgumentTypeError(f"Invalid argument pair {arg_pairs}, must be key=value")
    return tuple(arg_pairs.split("="))


def main(args):
    if not args.directory.exists():
        raise FileNotFoundError(f"Input directory {args.directory} not found")

    # Check if input is a single file or a directory of files.
    if args.directory.is_file():
        files = [args.directory]
    else:
        if args.mode == "json":
            suffix = ".json"
        elif args.mode in ["pointcloud", "heatmap3d"]:
            suffix = ".txt"
        else:
            suffix = ".pkl"
        files = sorted(list(args.directory.glob("*" + suffix)))
        logger.debug(f"Found {len(files)} files in {args.directory}")
        if len(files) == 0:
            return []

    # Limit the number of frames to display.
    if args.n is not None:
        files = files[: args.n]
    # Downsample the frames.
    if args.downsample < 0 or args.downsample > len(files):
        logger.error(f"Invalid downsample factor {args.downsample}, using 1")
        args.downsample = 1
    if args.downsample > 1:
        files = files[:: args.downsample]

    # Define the iterator.
    file_iterator = iter(files)
    if args.show_progress:
        import tqdm

        file_iterator = tqdm.tqdm(files, total=len(files))

    # Load and process files into scenes.
    scenes = []
    if args.mode == "pointcloud":
        for file in file_iterator:
            point_cloud = np.loadtxt(file).reshape(-1, 6)
            scene = show_point_cloud(point_cloud[:, :3], colors=point_cloud[:, 3:])
            scenes.append({"scene": scene})

    elif args.mode == "heatmap3d":
        from matplotlib import colormaps
        import matplotlib.colors as mcolors

        for file in file_iterator:
            logger.debug(f"Loading file {file}")
            heatmap3d = np.loadtxt(file).reshape(-1, 4)

            values = heatmap3d[:, 3]
            norm = mcolors.Normalize(vmin=min(values), vmax=max(values))
            cmap = colormaps["jet"]
            colors = np.array([cmap(norm(value))[:3] for value in values])
            colors = (colors * 255).astype(np.uint8)

            scene = show_point_cloud(heatmap3d[:, :3], colors=colors)
            scenes.append({"scene": scene})

    elif args.mode == "scenes":
        for k, file in enumerate(file_iterator):
            with open(file, "rb") as f:
                scene_dict = pkl.load(f)
            scenes.append(scene_dict)

    elif args.mode == "json":
        for k, file in enumerate(file_iterator):
            kwargs = {}
            if args.frame_meshes:
                kwargs["frame_meshes"] = tuple(args.frame_meshes)  # otherwise unhashable
            if args.smpl_model:
                kwargs["smpl_model"] = args.smpl_model
            scene_dict = load_scene_from_json(file, **kwargs)
            scenes.append(scene_dict)

    else:
        raise ValueError(f"Invalid displaying mode {args.mode}")

    # Add ground plane if requested. Only add it for the 'scene' tag in the scene_dict.
    # If the scene_dict does not have a 'scene' tag, the ground plane will not be added.
    if args.add_ground_plane:
        for scene_dict in scenes:
            if "scene" not in scene_dict:
                logger.warning("No scene found in scene_dict, not adding ground plane.")
                continue
            xy_min = -args.ground_plane_xy
            xy_max = args.ground_plane_xy
            z_g = args.ground_plane_z
            scene_dict["scene"] = show_grid(scene=scene_dict["scene"], z=z_g, xy_min=xy_min, xy_max=xy_max)

    return scenes


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("mode", choices=["pointcloud", "heatmap3d", "scenes", "json"])
    parser.add_argument("directory", type=pathlib.Path, help="Data directory or file")
    parser.add_argument("--fps", type=int, help="displaying fps", default=10)
    parser.add_argument(
        "--use-scene-cam",
        action="store_true",
        help="Use camera transform from trimesh scene," "if available.",
    )
    parser.add_argument(
        "--add-ground-plane", action="store_true", help="Add ground plane (z = 0) to the scene."
    )
    parser.add_argument("--show-frame-index", action="store_true", help="Show frame index.")
    parser.add_argument("--show-progress", action="store_true", help="Show progress bar.")
    parser.add_argument("--n", type=int, help="Number of frames to display", default=None)
    parser.add_argument("--downsample", type=int, help="Downsample factor", default=1)
    parser.add_argument(
        "--smpl-model",
        type=pathlib.Path,
        help="SMPL model for json processing, if needed.",
        default=None,
    )
    parser.add_argument("--ground-plane-xy", type=float, help="Ground plane xy size", default=10)
    parser.add_argument("--ground-plane-z", type=float, help="Ground plane height", default=0.0)
    parser.add_argument(
        "--frame-meshes",
        nargs="+",
        type=parse_arg_pairs,
        help="Mesh to display in addition to a frame with key, format key=mesh-path/predefined-mesh-id."
        "Predefined mesh ids are [cam, drone].",
        default=(),
    )
    parser.add_argument("--debug", action="store_true", help="debug mode")
    args_ = parser.parse_args()

    # Setup logger according to debug mode.
    logger.remove()
    if args_.debug:
        logger.add(sys.stderr, level="DEBUG")
    else:
        logger.add(sys.stderr, level="INFO")

    # Construct and display scenes.
    display_dataset(
        main(args_),
        fps=args_.fps,
        use_scene_cam=args_.use_scene_cam,
        debug=args_.debug,
        show_frame_index=args_.show_frame_index,
    )



import base64
import struct
import numpy as np
import json
import traceback
import trimesh

from functools import lru_cache
from loguru import logger
from pathlib import Path
from scipy.spatial.transform import Rotation
from typing import Dict, List, Optional, Tuple

from balanna.components import (
    show_trajectory,
    show_axis,
    show_capsule,
    show_cylinder,
    show_ellipsoid,
    show_mesh,
    show_trimesh,
    show_plane,
    show_sphere,
    show_point_cloud,
    RGBorRGBAType,
)


def __parse_colors(
    json_dict: Dict[str, List], key: str = "color", default: Tuple[int, int, int] = (255, 0, 0)
) -> RGBorRGBAType:
    if key not in json_dict:
        logger.debug(f"Color not found, using default color {default} using key {key}")
        return default
    if len(json_dict[key]) not in [3, 4]:
        logger.debug(f"Invalid color {json_dict[key]}, using default color {default}")
        return default

    red = int(json_dict[key][0] * 255)
    green = int(json_dict[key][1] * 255)
    blue = int(json_dict[key][2] * 255)

    if len(json_dict[key]) == 4:
        alpha = int(json_dict[key][3] * 255)
        return red, green, blue, alpha
    return red, green, blue


def __parse_position(json_dict: Dict[str, List], key: str = "position") -> Optional[np.ndarray]:
    if key not in json_dict.keys():
        logger.debug(f"Invalid position element, `{key}` not found")
        return None
    position = np.array(json_dict[key])
    if len(position) != 3:
        logger.debug(f"Invalid position element, got invalid `{key}` array {position}")
        return None
    return position


def __parse_pose(json_dict: Dict[str, List]) -> Optional[np.ndarray]:
    position = __parse_position(json_dict)
    if position is None:
        return None

    if "orientation" in json_dict.keys():
        orientation = np.array(json_dict["orientation"])
        if len(orientation) != 4:
            logger.debug(
                f"Invalid poses element, got invalid `orientation` array {orientation}, "
                f"expected quaternion"
            )
            return None
    elif "orientationRPY" in json_dict.keys():
        orientation_rpy = np.array(json_dict["orientationRPY"])
        if len(orientation_rpy) != 3:
            logger.debug(
                f"Invalid poses element, got invalid `orientationRPY` array {orientation_rpy}, "
                f"expected roll-pitch-yaw"
            )
            return None
        orientation = Rotation.from_euler("XYZ", orientation_rpy, degrees=False).as_quat()
    else:
        orientation = np.array([0, 0, 0, 1], dtype=np.float32)

    return np.concatenate((position, orientation))


def __parse_transform(json_dict: Dict[str, List]) -> Optional[np.ndarray]:
    pose = __parse_pose(json_dict)
    if pose is None:
        return None

    transform = np.eye(4)
    transform[:3, :3] = Rotation.from_quat(pose[3:]).as_matrix()
    transform[:3, 3] = pose[:3]
    return transform


def __parse_poses(json_dict: Dict[str, List], key: str = "poses") -> Optional[np.ndarray]:
    if key not in json_dict:
        logger.debug(f"Array of positions not found in {json_dict} using key {key}")
        return None
    if len(json_dict[key]) == 0:
        logger.debug(f"Array of positions is empty in {json_dict} using key {key}")
        return None

    num_poses = len(json_dict[key])
    poses = np.zeros((num_poses, 7), dtype=np.float32)
    for k, pose in enumerate(json_dict[key]):
        if not isinstance(pose, dict):
            logger.debug(f"Invalid type of poses elements, should be dicts, got {type(pose)}")
            return None
        poses[k] = __parse_pose(pose)
    return poses


def __parse_radii(json_dict: Dict[str, List], key: str = "radii") -> Optional[np.ndarray]:
    if key not in json_dict:
        logger.debug(f"Radii not found in {json_dict} using key {key}")
        return None
    radii = np.array(json_dict[key])
    if len(radii) != 3:
        logger.debug(f"Invalid radii array {radii} in {json_dict} using key {key}")
        return None
    return radii


def __parse_lines_count(json_dict: Dict[str, List]) -> Optional[Tuple[int, int]]:
    latitude = json_dict.get("countLatitude", None)
    longitude = json_dict.get("countLongitude", None)
    if latitude is None or longitude is None:
        logger.debug("Count not found, returning None")
        return None
    if not isinstance(latitude, int) or not isinstance(longitude, int):
        logger.debug(f"Invalid count {latitude, longitude}, must be integers, returning None")
        return None
    if latitude < 3 or longitude < 3:
        logger.debug(f"Invalid count {latitude, longitude}, must be >= 3, returning None")
        return None
    return latitude, longitude


def __parse_vertex_colors(json_dict: Dict[str, List], num_vertices: int) -> Optional[np.ndarray]:
    colors = json_dict.get("vertex_colors", None)
    if colors is not None:
        colors = np.array(colors) * 255
    elif "color" in json_dict:
        colors = np.array(json_dict["color"]) * 255
        colors = np.tile(colors, (num_vertices, 1))
    else:
        colors = None
    return colors


@lru_cache(maxsize=1)
def __process_kwargs(**kwargs):
    output = dict()
    if "smpl_model" in kwargs.keys():
        smpl_model_path = kwargs["smpl_model"]
        logger.debug(f"Loading SMPL model from {smpl_model_path}")
        import torch
        import smplx

        # device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        device = torch.device("cpu")
        output["smpl_model"] = smplx.SMPL(smpl_model_path, batch_size=1).to(device)
        output["device"] = device

    if "frame_meshes" in kwargs.keys():
        frame_meshes = kwargs["frame_meshes"]
        logger.info(f"Loading frame meshes from {frame_meshes}")
        meshes_dict = {}
        for key, path in frame_meshes:
            if path in ["cam", "drone"]:  # predefined meshes
                logger.debug(f"Loading predefined mesh {path}")
                path = Path(__file__).parent.parent / "meshes" / f"{path}.obj"
            logger.debug(f"Loading mesh {key} from {path}")
            try:
                meshes_dict[key] = trimesh.load_mesh(path)
            except FileNotFoundError:
                logger.error(f"Error loading frame mesh {key} from {path}")
                logger.debug(traceback.format_exc())
        output["frame_meshes"] = meshes_dict

    return output


def load_scene_from_json(file_path: Path, **kwargs):
    with open(file_path) as f:
        data = json.load(f)

    # Prepare the kwargs, if any.
    tools = __process_kwargs(**kwargs)

    # Process the data. Iterate through the json objects and call the respective visualization
    # functions, depending on the object type.
    scene_dict = dict()
    scene = trimesh.Scene()
    for name, values in data.items():
        if "type" not in values:
            logger.warning(f"Invalid object {name}, no type found, skipping")
            continue
        object_type = values["type"]
        logger.debug(f"Processing object {name} with type {object_type}")

        try:
            if object_type == "trajectory":
                positions = __parse_poses(values, "poses")[:, :3]
                if positions is None:
                    continue
                color = __parse_colors(values, "color")
                scene = show_trajectory(positions, colors=color, scene=scene)

            elif object_type == "frame":
                transform = __parse_transform(values)
                if transform is None:
                    continue
                scene = show_axis(transform, name=name, scene=scene)

                # If requested, add a mesh to the frame.
                if "frame_meshes" in tools and name in tools["frame_meshes"]:
                    mesh = tools["frame_meshes"][name]
                    scene = show_trimesh(mesh, name=name, scene=scene, transform=transform)

            elif object_type == "capsule":
                radius = values.get("radius", None)
                if radius is None:
                    logger.warning(f"Invalid capsule object {name}, no radius found, skipping")
                    continue
                p1 = __parse_position(values, "p1")
                p2 = __parse_position(values, "p2")
                if p1 is None or p2 is None:
                    logger.warning(f"Invalid capsule object {name}, no p1/p2 found, skipping")
                    continue
                color = __parse_colors(values, "color")
                count = __parse_lines_count(values)
                scene = show_capsule(p1, p2, radius, color=color, scene=scene, count=count)

            elif object_type == "sphere":
                radius = values.get("radius", None)
                if radius is None:
                    logger.warning(f"Invalid sphere object {name}, no radius found, skipping")
                    continue
                center = __parse_position(values, "center")
                if center is None:
                    logger.warning(f"Invalid sphere object {name}, no center found, skipping")
                    continue
                count = __parse_lines_count(values)
                color = __parse_colors(values, "color")
                scene = show_sphere(center, radius, color=color, scene=scene, count=count)

            elif object_type == "ellipsoid":
                radii = __parse_radii(values, "radii")
                if radii is None:
                    logger.warning(f"Invalid ellipsoid object {name}, no radii found, skipping")
                    continue
                center = __parse_position(values, "center")
                if center is None:
                    logger.warning(f"Invalid ellipsoid object {name}, no center found, skipping")
                    continue
                color = __parse_colors(values, "color")
                count = __parse_lines_count(values)
                scene = show_ellipsoid(center, radii, color=color, scene=scene, count=count)

            elif object_type == "cylinder":
                radius = values.get("radius", None)
                if radius is None:
                    logger.warning(f"Invalid cylinder object {name}, no radius found, skipping")
                    continue
                height = values.get("height", None)
                if height is None:
                    logger.warning(f"Invalid cylinder object {name}, no height found, skipping")
                    continue
                transform = __parse_transform(values)
                if transform is None:
                    logger.warning(
                        f"Invalid cylinder object {name}, no transform found, using identity"
                    )
                    transform = np.eye(4)
                color = __parse_colors(values, "color")
                count = values.get("count", None)
                scene = show_cylinder(
                    transform, radius, height, color=color, scene=scene, count=count
                )

            elif object_type == "point_cloud":
                points = values.get("points", None)
                if points is None:
                    logger.warning(f"Invalid point cloud object {name}, no points found, skipping")
                    continue

                # Decode base64 encoded points if base64 encoded.
                if isinstance(points, str):
                    points_decoded = base64.b64decode(points)
                    num_points = len(points_decoded) // 4
                    points = list(struct.unpack("f" * num_points, points_decoded))
                # If points are not base64 encoded, check if they are a list of lists.
                elif isinstance(points, list) or isinstance(points, np.ndarray):
                    pass
                else:
                    logger.warning(
                        f"Invalid point cloud object {name}, "
                        "points must be a list or base64 encoded, skipping"
                    )
                    continue

                points = np.array(points)
                if len(points) == 0:
                    logger.warning(f"Invalid point cloud object {name}, no points found, skipping")
                    continue

                colors = np.array(values["colors"]) * 255 if "colors" in values else None
                point_size = values.get("point_size", 4.0)
                if point_size < 0.1:
                    logger.debug(
                        f"Point size {point_size} of point cloud too small, multiplying by 500"
                    )
                    point_size *= 500
                scene = show_point_cloud(points, colors=colors, point_size=point_size, scene=scene)

            elif object_type == "mesh":
                vertices = values.get("vertices", None)
                if vertices is None:
                    logger.warning(f"Invalid mesh object {name}, no vertices found, skipping")
                    continue
                vertices = np.array(vertices)

                faces = values.get("faces", None)
                if faces is None:
                    logger.warning(f"Invalid mesh object {name}, no faces found, skipping")
                    continue
                faces = np.array(faces)

                colors = __parse_vertex_colors(values, vertices.shape[0])
                scene = show_mesh(vertices, faces, vertex_color=colors, scene=scene)

            elif object_type == "smpl_mesh":
                import torch

                if "device" not in tools or "smpl_model" not in tools:
                    logger.warning(f"SMPL model not loaded, skipping SMPL object {name}")
                    continue
                device = tools["device"]
                smpl_model = tools["smpl_model"]

                if "betas" not in values:
                    betas = torch.zeros(10, device=device)
                else:
                    betas = torch.tensor(values["betas"], device=device)
                    if len(betas) != 10:
                        logger.warning(
                            f"Invalid SMPL object {name}, betas must be of length 10, skipping"
                        )
                        continue
                betas = betas.unsqueeze(0)

                if "thetas" not in values:
                    logger.warning(f"Invalid SMPL object {name}, no thetas found, skipping")
                    continue
                thetas = torch.tensor(values["thetas"], device=device)
                if len(thetas) != 72:
                    logger.warning(
                        f"Invalid SMPL object {name}, thetas must be of length 72, skipping"
                    )
                    continue
                thetas = thetas.unsqueeze(0)

                if "translation" not in values:
                    translation = torch.zeros(3, device=device)
                else:
                    translation = torch.tensor(values["translation"], device=device)
                    if len(translation) != 3:
                        logger.warning(
                            f"Invalid SMPL object {name}, translation must be of length 3, skipping"
                        )
                        continue
                translation = translation.unsqueeze(0)

                vertices = smpl_model.forward(
                    betas, global_orient=thetas[:, :3], body_pose=thetas[:, 3:]
                ).vertices
                vertices = vertices + translation
                vertices = vertices[0].detach().cpu().numpy()

                colors = __parse_vertex_colors(values, 6890)
                scene = show_mesh(vertices, smpl_model.faces, vertex_color=colors, scene=scene)

            elif object_type == "message":
                text = str({k: v for k, v in values.items() if k != "type"})
                scene_dict[name] = text

            elif object_type == "plane":
                if "normal" not in values:
                    logger.warning(f"Invalid plane object {name}, no normal found, skipping")
                    continue
                if "center" not in values:
                    logger.warning(f"Invalid plane object {name}, no center found, skipping")
                    continue
                normal = np.array(values["normal"])
                center = np.array(values["center"])
                extent_xy = values.get("extent_xy", 5.0)
                extent_z = values.get("extent_z", 0.01)
                color = __parse_colors(values, "color")
                scene = show_plane(normal, center, extent_xy, extent_z, color=color, scene=scene)

            else:
                logger.warning(f"Invalid object {name}, unknown type {object_type}, skipping")
                continue

        except Exception as e:
            logger.error(f"Error processing object {name}, continuing")
            logger.debug(traceback.format_exc())
            continue

    if "scene" in scene_dict:
        logger.warning(f"Scene dict already has a `scene` key, overwriting")
    scene_dict["scene"] = scene
    return scene_dict
